<?php

namespace Shavshukov\EventAppBase\Console;

use Shavshukov\EventAppBase\Service\ErrorReport;
use Shavshukov\EventAppBase\ApplicationInterface;
use React\Promise;

abstract class Kernel
{
    /**
     * Application instance
     *
     * @var \Shavshukov\EventAppBase\ApplicationInterface
     */
    protected $app;
    
    /**
     * Array of arguments passed to the routed action
     *
     * @var array|null
     */
    protected $argv;
    
    /**
     * File name of the entry point file
     *
     * @var string
     */
    protected $entryFileName;
    
    /**
     * Action name
     *
     * @var string
     */
    protected $action;
    
    /**
     * Commands array
     *
     * @var array
     */
    protected $commands;
    
    /**
     * Constructor
     * 
     * @param \Shavshukov\EventAppBase\ApplicationInterface $app
     */
    public function __construct(ApplicationInterface $app)
    {
        $this->app = $app;
        $this->argv = $_SERVER['argv'] ?? null;
        $this->entryFileName = $this->argv[0] ?? '???';
        $this->action = ($this->argv[1] ?? null) ?: 'help';
        $this->commands = $this->getCommands();
    }
    
    /**
     * Get classes of commands available for the application
     * 
     * @return string[]
     */
    protected abstract function getCommands(): array;
    
    /**
     * Find command, launch it with arguments, start application loop
     * 
     * @return null
     */
    public function handle()
    {
        if (isset($this->commands[$this->action])) {
            $command = $this->createCommand($this->commands[$this->action]);
            if (empty($this->argv)) {
                $promise = Promise\resolve($command->handle([]));
            } else {
                $promise = Promise\resolve($command->handle(array_slice($this->argv, 2)));
            }
        } else {
            if (!$this->action || $this->action === 'help') {
                if (isset($this->argv[2])) {
                    $this->displayHelp($this->argv[2]);
                } else {
                    $this->displayUsage();
                }
            } else {
                echo "Unknown command: {$this->action} \n";
            }
            
            $promise = Promise\resolve();
        }
        
        $promise->then(null, function (\Throwable $e) {
            ErrorReport::report($e, true);
            exit(-1);
        });
        
        $this->app->run();
    }
    
    /**
     * Display usage information
     * 
     * @return null
     */
    public function displayUsage()
    {
        $usage = [];
        
        // describe application
        if (($description = $this->app->description())) {
            $usage[] = $description;
        }
        
        $usage[] = "Usage: {$this->entryFileName} <command> [<args>]\n";
        $usage[] = "{$this->entryFileName} {$this->padCommandName('help')} Display help information";
        
        // add commands usage info
        foreach ($this->commands as $commandName => $commandClass) {
            $usage[] = "{$this->entryFileName} {$this->padCommandName($commandName)} {$this->createCommand($commandClass)->usage()}";
        }
        
        $usage[] = '';
        $usage[] = "See '{$this->entryFileName} help <command>'";
        
        echo implode("\n", $usage);
    }
    
    /**
     * Display command help information
     * 
     * @param string $commandName
     * @return null
     */
    public function displayHelp(string $commandName)
    {
        $help = [];
        
        // describe application
        if (($description = $this->app->description())) {
            $help[] = $description;
        }
        
        if (isset($this->commands[$commandName])) {
            $command = $this->createCommand($this->commands[$commandName]);
            $help[] = $command->help($this->entryFileName);
        } else {
            $help[] = "No such command: {$commandName}\n";
            $help[] = "See '{$this->entryFileName} help' for the list of commands";
        }
        
        echo implode("\n", $help);
        
    }
    
    /**
     * Get padded command name
     * 
     * @param string $commandName
     * @return string
     */
    protected function padCommandName(string $commandName): string
    {
        return str_pad($commandName, 15, ' ');
    }
    
    /**
     * Create command from class
     * 
     * @param string $commandClass
     * @return \Shavshukov\EventAppBase\Console\CommandInterface
     */
    protected function createCommand(string $commandClass): CommandInterface
    {
        $commandInstance = new $commandClass();
        
        if (!$commandInstance instanceof CommandInterface) {
            throw new \Exception("Every command must be an instance of \Shavshukov\EventAppBase\Console\CommandInterface, but {$commandClass} is not");
        }
        
        $commandInstance->setApplication($this->app);
        
        return $commandInstance;
    }
}