<?php

namespace Shavshukov\EventAppBase\Http;

use Shavshukov\EventAppBase\Service\ErrorReport;
use Shavshukov\EventAppBase\ApplicationInterface;

use Psr\Http\Message\ServerRequestInterface;
use React;

abstract class Server
{
    /**
     * Instance of the console application running this HTTP server
     *
     * @var \Shavshukov\EventAppBase\ApplicationInterface
     */
    protected $app;
    
    /**
     * Router instance
     *
     * @var \Shavshukov\EventAppBase\Http\Routing\Router
     */
    protected $router;
    
    /**
     * Path to the public folder
     *
     * @var string
     */
    protected $publicPath;
    
    /**
     * Real path to the public folder
     *
     * @var string|null
     */
    protected $publicRealPath;
    
    /**
     * Constructor
     * 
     * @param \Shavshukov\EventAppBase\ApplicationInterface $app
     */
    public function __construct(ApplicationInterface $app)
    {
        $this->app = $app;
        $this->router = $this->getRouterInstance();
        $this->publicPath = $app->getAppPath() . DIRECTORY_SEPARATOR . 'public';
        $this->publicRealPath = realpath($this->publicPath);
    }
    
    /**
     * Start serving the HTTP server
     * 
     * @return null
     */
    public function serve()
    {
        $httpPort = $this->app->config('http_port');
        $httpListen = $this->app->config('http_listen', '0.0.0.0');
        $bind = "{$httpListen}:{$httpPort}";

        $httpServer = new React\Http\Server(function (ServerRequestInterface $request) {
            return $this->router->processRequest($request);
        });

        $socket = new React\Socket\Server($bind, $this->app->getLoop());
        
        if ($this->app->config('http_https')) {
            $socket = new React\Socket\SecureServer($socket, $this->app->getLoop(), [
                'local_cert' => $this->app->config('http_https_cert')
            ]);
        }
        
        $httpServer->listen($socket);

        $httpServer->on('error', function (\Exception $e) {
            ErrorReport::report($e, false);
        });

        echo "Server running at http://{$bind}\n";
    }

    /**
     * Get path to the public folder
     * 
     * @return string
     */
    public function getPublicPath(): string
    {
        return $this->publicPath;
    }
    
    /**
     * Get real path to the public folder
     * 
     * @return string
     */
    public function getPublicRealPath(): string
    {
        return $this->publicRealPath;
    }
    
    /**
     * Get new HTTP router instance
     * 
     * @return \Shavshukov\EventAppBase\Http\Routing\Router
     */
    protected abstract function getRouterInstance(): Routing\Router;
    
}
