<?php

namespace Shavshukov\EventAppBase\Model;

use Shavshukov\EventAppBase\Interfaces\ModelInterface;

trait ModelDataTrait
{
    /**
     * Array of data
     *
     * @var array
     */
    protected $data;
    
    /**
     * Is array of data changed since load
     *
     * @var bool
     */
    protected $modified = false;
    
    /**
     * Factory method to create a model from another model
     * 
     * @param \Shavshukov\EventAppBase\Interfaces\ModelInterface $model
     * @return \static
     */
    public static function copyOf(ModelInterface $model)
    {
        return (new static())->setDataFrom($model);
    }
    
    /**
     * Check that value at path exists
     * 
     * @param string $path
     * @return bool
     */
    public function has(string $path): bool
    {
        $array = $this->data;
        
        foreach (explode('.', $path) as $segment) {
            if (is_array($array) && array_key_exists($segment, $array)) {
                $array = $array[$segment];
            } else {
                return false;
            }
        }

        return true;
    }
    
    /**
     * Get value at path
     * 
     * @param string $path
     * @param mixed $default
     * @return mixed
     */
    public function get(string $path, $default = null)
    {
        $array = $this->data;
        
        foreach (explode('.', $path) as $segment) {
            if (is_array($array) && array_key_exists($segment, $array)) {
                $array = $array[$segment];
            } else {
                return $default;
            }
        }

        return $array;
    }
    
    /**
     * Get underlying array of the model
     * 
     * @return array
     */
    public function all(): array
    {
        return $this->data;
    }
    
    /**
     * Modify data with new data
     * 
     * @param array $modification
     * @return $this
     */
    public function modify(array $modification)
    {
        $this->data = array_replace_recursive($this->data, $modification);
        $this->modified = true;
        
        return $this;
    }

    /**
     * Copy data from another model
     * 
     * @param \Shavshukov\EventAppBase\Interfaces\ModelInterface $model
     * @return \static
     */
    public function setDataFrom(ModelInterface $model)
    {
        $this->data = $model->all();
        $this->modified = true;
        
        return $this;
    }
    
}
