<?php

namespace Shavshukov\EventAppBase\Service;

use React\Promise;

class Container
{
    /**
     * Contained service instances\closures
     *
     * @var array
     */
    protected $services = [];
    
    /**
     * Whether the contained name is a closure
     *
     * @var bool[]
     */
    protected $closures = [];
    
    /**
     * Whether the contained name is a singleton
     *
     * @var bool[]
     */
    protected $singletons = [];
    
    /**
     * Instances for resolves
     *
     * @var array
     */
    protected $instances = [];
    
    /**
     * Register given service providers
     * 
     * @param array $serviceProviders
     * @param array $registerArguments
     * @throws \Exception
     */
    public function registerProviders(array $serviceProviders, ...$registerArguments)
    {
        foreach ($serviceProviders as $serviceProvider) {
            if (is_string($serviceProvider)) {
                $providerInstance = new $serviceProvider();
            } else {
                $providerInstance = $serviceProvider;
            }
            
            if (!$providerInstance instanceof ProviderInterface) {
                throw new \Exception('Every service provider must be an instance of ' . Provider::class);
            }
            
            $providerInstance->register($this, ...$registerArguments);
        }
    }
    
    /**
     * Get whether a service is bound to a name
     * 
     * @param string $name
     * @return bool
     */
    public function bound(string $name): bool
    {
        return isset($this->services[$name]);
    }
    
    /**
     * Bind a service to the container
     * 
     * @param string $name
     * @param \Closure|mixed $closureOrInstance
     * @param bool $singleton
     * @throws \Exception
     */
    public function bind(string $name, $closureOrInstance, bool $singleton = false)
    {
        if ($this->singletons[$name] ?? false) {
            throw new \Exception("Cannot rebind service: {$name} is a singleton");
        }
        
        $this->services[$name] = $closureOrInstance;
        $this->closures[$name] = $closureOrInstance instanceof \Closure;
        $this->singletons[$name] = $singleton;
        unset($this->instances[$name]);
    }
    
    /**
     * Resolve a service from the container
     * 
     * @param string $name
     * @param array $serviceArguments
     * @return \React\Promise\PromiseInterface
     */
    public function resolve(string $name, ...$serviceArguments): Promise\PromiseInterface
    {
        if (!isset($this->services[$name])) {
            return Promise\reject(new \Exception("Not found binding for service: {$name}"));
        }
        
        if (isset($this->instances[$name])) {
            return Promise\resolve($this->instances[$name]);
        }
        
        if ($this->closures[$name]) {
            return Promise\resolve($this->instances[$name] = $this->services[$name](...$serviceArguments));
        } else {
            return Promise\resolve($this->instances[$name] = $this->services[$name]);
        }
    }
    
    /**
     * Unbind a service from the container
     * 
     * @param string $name
     * @return null
     */
    public function unbind(string $name)
    {
        unset($this->services[$name], $this->closures[$name], $this->singletons[$name], $this->instances[$name]);
    }
}
