<?php

namespace Shavshukov\EventAppBase\Storage;

use Shavshukov\EventAppBase\FileReaders\FileReaderInterface;
use Shavshukov\EventAppBase\Interfaces\StorageInterface;

use React\Promise;

class FileStorage extends Storage
{
    /**
     * File reader instance
     *
     * @var \Shavshukov\EventAppBase\FileReaders\FileReaderInterface
     */
    protected $fileReader;
    
    /**
     * File to be read from and written to
     *
     * @var string
     */
    public $fileName;
    
    /**
     * Necessity of the file to exist when the storage is loaded
     *
     * @var bool
     */
    public $mustExist = false;
   
    /**
     * Constructor
     * 
     * @param \Shavshukov\EventAppBase\FileReaders\FileReaderInterface $fileReader
     * @param string $fileName
     */
    public function __construct(FileReaderInterface $fileReader, string $fileName)
    {
        $this->fileReader = $fileReader;
        $this->fileName = $fileName;
    }
    
    /**
     * {@inheritDoc}
     */
    public function setBackEnd(StorageInterface $backend)
    {
        throw new \Exception('Backends are not supported for FileStorage');
    }
    
    /**
     * Get file name
     * 
     * @return string
     */
    public function getFileName(): string
    {
        return $this->fileName;
    }

    /**
     * {@inheritDoc}
     */
    public function content()
    {
        return $this->source;
    }
    
    /**
     * {@inheritDoc}
     */
    public function load(): bool
    {
        $fileName = $this->getFileName();

        if (file_exists($fileName)) {
            $this->source = file_get_contents($fileName);
            $this->modified = false;
            return true;
        }
        
        if ($this->mustExist) {
            throw new \Exception("File does not exists: {$fileName}");
        }
        
        if ($this->source) {
            $this->source = null;
            $this->modified = false;
            return true;
        }

        return false;
    }
    
    /**
     * {@inheritDoc}
     */
    public function modifyBy(StorageInterface $storage): bool
    {
        $source = $storage->source();
        
        if (!is_string($source)) {
            throw new \Exception('Provided incompatible source: expected string data for FileStorage');
        }

        $this->modified = ($this->source !== $source);
        $this->source = $source;
        return $this->save();
    }
    
    /**
     * {@inheritDoc}
     */
    public function loadAsync(): Promise\PromiseInterface
    {
        $fileName = $this->getFileName();
        
        if (file_exists($fileName)) {
            return $this->fileReader->getContents($fileName)->then(function (string $contents) {
                $this->source = $contents;
                $this->modified = false;
                return true;
            });
        }
        
        if ($this->mustExist) {
            return Promise\reject(new \Exception("File does not exists: {$fileName}"));
        }
        
        if ($this->source) {
            $this->source = null;
            $this->modified = false;
            return Promise\resolve(true);
        }
        
        return Promise\resolve(false);
    }
    
    /**
     * {@inheritDoc}
     */
    public function save(): bool
    {
        if ($this->modified) {
            $fileName = $this->getFileName();
            file_put_contents($fileName, $this->source);
            
            return true;
        }
        
        return false;
    }

}
